import bhi160
import display
import utime
from urandom import randint


class Vector:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        return 'Vector(x={}, y={})'.format(self.x, self.y)


class Target:
    def __init__(self, x, y, radius):
        self.x = x
        self.y = y
        self.radius = radius

    def __contains__(self, other):
        if type(other) != Vector:
            raise('it is only possible to check whether a Vector is'
                  'within Target')
        return (
            (self.x - self.radius < other.x < self.x + self.radius)
            and
            (self.y - self.radius < other.y < self.y + self.radius)
        )

    def __repr__(self):
        return 'Target(x={}, y={}, radius={})'.format(
                self.x, self.y, self.radius)


class Timing:
    def __init__(self, timeout=15, start=None):
        self._timeout = timeout
        self._timestamp = start if start is not None else utime.time_ms()

    def malus(self, malus=0.5):
        self._timeout -= malus

    def remaining_time(self):
        print('timestamp', self._timestamp)
        print('timeout', self._timeout)
        now = utime.time_ms()
        print('now', now)
        remaining_time = (float(self._timeout)
                          - float(now - self._timestamp) / 1000)
        print('remaining time', remaining_time)
        if remaining_time < 0:
            remaining_time = 0.0
        return remaining_time

    def draw(self, disp, posx=115, posy=15):
        disp.print('{:.2f}'.format(self.remaining_time()), posx=posx,
                   posy=posy, font=display.FONT8)


class Credits:
    def __init__(self):
        self.credit = 0

    def add_score(self, remaining_time):
        self.credit += round(remaining_time)

    def draw(self, disp, posx=5, posy=5):
        disp.print('Points: {: 4d}'.format(self.credit), posx=posx,
                   posy=posy, font=display.FONT8)


class Level:
    def __init__(self):
        self.level = 1

    def update_level(self, credits):
        level = credits.credit // 20 + 1
        if level > self.level:
            self.level = level
            return True
        else:
            return False

    def draw(self, disp, posx=115, posy=5):
        disp.print('lvl: {: 4d}'.format(self.level), posx=posx,
                   posy=posy, font=display.FONT8)


class Stats:
    def __init__(self, hits=0, level=0, circle_size=10, hits_per_level=5,
                 accel_factor=15, time_to_next_hit=15):
        self.hits = hits
        self.circle_size = circle_size
        self.accel_factor = accel_factor
        self.time_to_next_hit = time_to_next_hit
        self.malus = 0.5

    def __repr__(self):
        return (('Stats(hits={}, circle_size={}, '
                 'accel_factor={}, time_to_next_hit={})')
                .format(self.hits, self.circle_size,
                        self.accel_factor,
                        self.time_to_next_hit))

    def level_up(self):
        self.accel_factor = round(self.accel_factor * 1.5)
        if self.time_to_next_hit > 5:
            self.time_to_next_hit -= 1
        if self.malus < self.time_to_next_hit:
            self.malus += 0.5
        if self.circle_size > 4:
            self.circle_size -= 1

    def new_hit(self):
        self.hits += 1


stats = Stats()

DISPLAY = Vector(x=160, y=80)
accel_factor = 5
DECEL_FACTOR = 0.66
DELAY = 0.01

# circle
target = Target(x=randint(0, DISPLAY.x), y=randint(0, DISPLAY.y),
                radius=stats.circle_size)
credits = Credits()
timer = Timing(timeout=stats.time_to_next_hit)
level = Level()

bhi = bhi160.BHI160Accelerometer()

ball_origin = Vector(x=DISPLAY.x / 2, y=DISPLAY.y / 2)
ball_velocity = Vector(x=0, y=0)

ball_trace = list()
# setup display
with display.open() as d:
    if hasattr(d, 'backlight'):
        d.backlight(1000)


timestamp = utime.time_ms()
while True:
    while True:
        current_sample = bhi.read()
        if len(current_sample) > 0:
            current_sample = current_sample[0]
            break
        utime.sleep(DELAY)
    cumm_sleep_time = float(utime.time_ms() - timestamp) / 1000
    timestamp = utime.time_ms()
    print('current delay %f' % cumm_sleep_time)
    print('accel', current_sample)

    ball_velocity = Vector(
        x=ball_velocity.x
        + stats.accel_factor * cumm_sleep_time * current_sample.x,
        y=ball_velocity.y
        - stats.accel_factor * cumm_sleep_time * current_sample.y
    )
    print('ball_velocity', ball_velocity)

    ball_origin = Vector(
        x=ball_origin.x + cumm_sleep_time * ball_velocity.x,
        y=ball_origin.y + cumm_sleep_time * ball_velocity.y
    )

    print('ball_origin', ball_origin)

    if ball_origin.x > DISPLAY.x:
        ball_velocity.x *= - DECEL_FACTOR
        ball_velocity.y *= DECEL_FACTOR
        remainder = ball_origin.x - DISPLAY.x
        ball_origin.x = DISPLAY.x - remainder
        timer.malus()

    if ball_origin.x < 0:
        ball_velocity.x *= - DECEL_FACTOR
        ball_velocity.y *= DECEL_FACTOR
        ball_origin.x *= -1
        timer.malus()

    if ball_origin.y > DISPLAY.y:
        ball_velocity.x *= DECEL_FACTOR
        ball_velocity.y *= - DECEL_FACTOR
        remainder = ball_origin.y - DISPLAY.y
        ball_origin.y = DISPLAY.y - remainder
        timer.malus()

    if ball_origin.y < 0:
        ball_velocity.x *= DECEL_FACTOR
        ball_velocity.y *= - DECEL_FACTOR
        ball_origin.y *= -1
        timer.malus()

    if ball_origin in target:
        target = Target(x=randint(0, DISPLAY.x), y=randint(0, DISPLAY.y),
                        radius=stats.circle_size)
        credits.add_score(timer.remaining_time())
        if level.update_level(credits):
            stats.level_up()
        timer = Timing(timeout=stats.time_to_next_hit)
        stats.new_hit()

    print('ball_velocity', ball_velocity)

    next_elem = Vector(round(ball_origin.x), round(ball_origin.y))
    if len(ball_trace) == 20:
        ball_trace = ball_trace[1:] + [next_elem]
    else:
        ball_trace.append(next_elem)
    with display.open() as disp:
        disp.clear()

        # draw the circle
        disp.circ(target.x, target.y, target.radius, col=[255, 0, 0])
        disp.circ(target.x, target.y, target.radius - 2, col=[0, 0, 0])

        # draw the moving dot
        for i in range(len(ball_trace) - 1):
            col = round(255 / len(ball_trace) * i)
            disp.line(ball_trace[i].x, ball_trace[i].y,
                      ball_trace[i+1].x, ball_trace[i+1].y,
                      col=[col, col, col])

        credits.draw(disp)
        timer.draw(disp)
        level.draw(disp)

        disp.update()

    print('finished loop')
    utime.sleep(DELAY)